#!/bin/bash

AUTHOR="Janne Hellas"
VERSION="0.7"
LICENSE="Public Domain or equivalent."
LINFO="This software is provided \"as-is\" without any warranties or liabilities."

SCRIPTDIR=$(dirname "${BASH_SOURCE}")

# Files
if [[ ! -f "$SCRIPTDIR/codemap" ]]; then
   echo "$0: codemap file missing." 1>&2
   exit 1
fi

if [[ ! -f "$SCRIPTDIR/offsets" ]]; then
   echo "$0: offsets file missing." 1>&2
   exit 1
fi

if [[ ! -f "$SCRIPTDIR/config" ]]; then
   echo "$0: Config file not found. Creating.." 1>&2
   touch "$SCRIPTDIR/config"

   echo "# Text field separator" >> "$SCRIPTDIR/config"
   echo "TFS=00C8B5" >> "$SCRIPTDIR/config"
   echo "" >> "$SCRIPTDIR/config"
   echo "# Extraction (line number from offsets file)" >> "$SCRIPTDIR/config"
   echo "EXTRACT=2" >> "$SCRIPTDIR/config"
   echo "" >> "$SCRIPTDIR/config"
   echo "# Number of language translations and default language selection." >> "$SCRIPTDIR/config"
   echo "# Ie. how many TFS:s until the next language unit." >> "$SCRIPTDIR/config"
   echo "NBLAN=13" >> "$SCRIPTDIR/config"
   echo "LAN=1" >> "$SCRIPTDIR/config"
   echo "" >> "$SCRIPTDIR/config"
   echo "# Control character truncation" >> "$SCRIPTDIR/config"
   echo "# If set to 0, control characters will be replaced with values from codemap." >> "$SCRIPTDIR/config"
   echo "# If set to 1, control characters will be truncated." >> "$SCRIPTDIR/config"
   echo "TRUN=0" >> "$SCRIPTDIR/config"
   echo "" >> "$SCRIPTDIR/config"
   echo "# Input file location" >> "$SCRIPTDIR/config"
   echo "INPUTFILE=\"../FF6.exe\"" >> "$SCRIPTDIR/config"
   echo "" >> "$SCRIPTDIR/config"
fi
source "$SCRIPTDIR/config"

if [[ "$1" != "" ]]; then
   EXTRACT=$1
fi

if [[ "$2" != "" ]]; then
   LAN=$2
fi

if [[ "$3" != "" ]]; then
   TRUN=$3
fi

PERF=30
PERFOW=0
if [[ "$4" != "" ]]; then
   PERF=$4
   PERFOW=1
fi

if [[ "$5" != "" ]]; then
   INPUTFILE="$5"
fi

# Test environment
JS=`which bc`
if [[ "${JS}" == "" ]]; then
   echo "$0: Environmental failure. Install bc." 1>&2
   exit 1
fi

if [[ "$1" == "-h" ]] || [[ "$1" == "h" ]] || [[ "$1" == "-help" ]] || [[ "$1" == "help" ]]; then
   echo "CB Text Extractor v${VERSION}"
   echo "Extracts plain text data from a file."
   echo ""
   echo "Usage: ./cb_extractor.sh [feature] [lan] [trun] [performance] [Input file]"
   echo "feature is a line number from offsets file."
   echo "lan is a language selection."
   echo "trun is control character truncation [0-1]."
   echo "Performance is a numeric value."
   echo "Input file is the location of the file to extract from."
   echo ""
   echo "Parameters can be set to default values in config except performance which can be defined in offsets."
   echo "Control characters can be edited in codemap."
   echo ""
   echo "License: $LICENSE"
   echo "$LINFO"
   exit 0
fi

# Handle non-option arguments
if [[ $# -gt 5 ]]; then
   echo "$0: Too many parameters." 1>&2
   exit 3
fi

if [[ ! ${TFS} =~ ^[[:xdigit:]]+$ ]]; then
   echo "$0: Invalid text field separator. Define a hex number in config." 1>&2
   exit 2
fi

if [[ ! ${EXTRACT} =~ ^[0-9]+$ ]] || [[ $EXTRACT -lt 1 ]] || [[ $EXTRACT -gt `wc -l "$SCRIPTDIR/offsets" | cut -d' ' -f1` ]]; then
   echo "$0: Invalid extraction selection; specify a number in range." 1>&2
   exit 2
fi

if [[ ! ${NBLAN} =~ ^[0-9]+$ ]] || [[ $NBLAN -lt 0 ]] || [[ $NBLAN -gt 10000 ]]; then
   echo "$0: Invalid number of languages settings. Check config." 1>&2
   exit 2
fi

if [[ ! ${LAN} =~ ^[0-9]+$ ]] || [[ $LAN -lt 0 ]] || [[ $LAN -gt $(( $NBLAN - 1 )) ]]; then
   echo "$0: Invalid language selection; specify an integer [0-$(( $NBLAN - 1 ))]." 1>&2
   exit 2
fi

if [[ ! -f "$INPUTFILE" ]]; then
   echo "$0: Input file doesn't exist." 1>&2
   exit 2
fi

if [[ ! ${PERF} =~ ^[0-9]+$ ]] || [[ $PERF -lt 3 ]] || [[ $PERF -gt 1000 ]]; then
   echo "$0: Invalid performance variable; specify an integer [3-1000]." 1>&2
   exit 2
fi

if [[ ! ${TRUN} =~ ^[0-9]+$ ]] || [[ ${TRUN} -gt 1 ]]; then
   echo "$0: Invalid truncation setting; specify either 0 or 1."
   exit 2
fi

# Generate sed
SED=""
while read -r line
do
    if [[ "$line" == "" ]] || [[ "$line" =~ "#" ]]; then
       continue
    fi
    CHAR=${line%%:*}
    #CHAR=`echo "$CHAR" | xxd -p | sed 's/./&\\\\x/2' | sed 's/^/\\\\x/; s/0a$//'`
    CONTROL=""
    if [[ ${TRUN} -ne 1 ]]; then
       CONTROL=${line##*:}
    fi
    SED=${SED}`echo -n "s/\x00${CHAR}/${CONTROL}/g; "`
done < "$SCRIPTDIR/codemap"

# Empty control character truncation
if [[ ${TRUN} -eq 1 ]]; then
   SED=${SED}`echo -n "s/\x00//g; "`
fi

# Debug
#echo "Replace Sed"
#echo "$SED"
#echo ""

MM=0
CHUNK=1
LANOW=0

TCOMMENT=`awk NR=="${EXTRACT}" "$SCRIPTDIR/offsets" | cut -d: -f1`
if [[ "$TCOMMENT" =~ ^"#" ]] || [[ "$TCOMMENT" == "" ]]; then
   echo "$0: Offset row is empty or contains a comment." 1>&2
   exit 5
fi
ENAME=`awk NR=="${EXTRACT}" "$SCRIPTDIR/offsets" | cut -d: -f2`
RPERF=`awk NR=="${EXTRACT}" "$SCRIPTDIR/offsets" | cut -d: -f3`
RCHUNK=`awk NR=="${EXTRACT}" "$SCRIPTDIR/offsets" | cut -d: -f4`
RMM=`awk NR=="${EXTRACT}" "$SCRIPTDIR/offsets" | cut -d: -f5`

if [[ $PERFOW -eq 0 ]] && [[ "$RPERF" != "" ]]; then
   if [[ ${RPERF} =~ ^[0-9]+$ ]] && [[ $RPERF -gt 3 ]] && [[ $RPERF -lt 1000 ]]; then
      PERF=$RPERF
   else
      echo "$0: Invalid performance defined in offsets. Using default $PERF." 1>&2
   fi
fi
#echo "Perf debug: $PERF, rperf: $RPERF, ow: $PERFOW"
PERF=$(( $PERF * 2 ))

if [[ "$RCHUNK" != "" ]]; then
   if [[ ${RCHUNK} =~ ^[0-9]+$ ]] && [[ $RCHUNK -gt 0 ]] && [[ $RCHUNK -lt 1000 ]]; then
      CHUNK=$RCHUNK
   else
      echo "$0: Invalid nb of chunks definition in offsets. Using default $CHUNK." 1>&2
   fi
fi
#echo "Chunk debug: $CHUNK, rchunk: $RCHUNK"

if [[ "$RMM" != "" ]]; then
   if [[ ${RMM} =~ ^[0-9]+$ ]] && [[ $RMM -ge 0 ]] && [[ $RMM -lt 100000 ]]; then
      MM=$RMM
   else
      echo "$0: Invalid start index in offsets. Using default $MM." 1>&2
   fi
fi
#echo "MM debug: $MM, rmm: $RMM"

if [[ "$ENAME" =~ "subchunk" ]]; then
  echo "Subchunk location selected. Continue? (y/n)"
     read OVERWRITE
  if [[ $OVERWRITE == "yes" ]] || [[ $OVERWRITE == "y" ]]; then
     echo "Will resume."
     if [[ $PERFOW -eq 0 ]]; then
        PERF=50
     fi
  else
     echo "Nothing to do. Exiting."
     exit 0
  fi
fi

echo "$ENAME"

while [[ $CHUNK -gt 0 ]]
do
   INDEX=0
   i=0
   START1=`awk NR=="${EXTRACT}" "$SCRIPTDIR/offsets" | cut -d: -f1`
   if [[ ! ${START1} =~ ^[[:xdigit:]]+$ ]]; then
      echo "$0: Offset row was invalid." 1>&2
      exit 5
   fi
   END1=`awk NR=="$(( ${EXTRACT} + 1 ))" "$SCRIPTDIR/offsets" | cut -d: -f1`
   if [[ ! ${END1} =~ ^[[:xdigit:]]+$ ]]; then
      echo "$0: End offset was invalid." 1>&2
      exit 5
   fi
   ROM_OFFSET=$(echo "obase=10; ibase=16; ${START1}" | bc)
   SIZE=$(( (0x$END1 - 0x$START1) + 3 ))
   DATAHEX=`xxd -p -l $SIZE -s $ROM_OFFSET -u "${INPUTFILE}" | tr -d " \n"`

   RAPPEND=`awk NR=="${EXTRACT}" "$SCRIPTDIR/offsets" | cut -d: -f6`
   if [[ "${RAPPEND}" =~ "astart" ]]; then
       DATAHEX=${TFS}${DATAHEX}
       SIZE=$(( $SIZE + ${#TFS} / 2 ))
   fi
   if [[ "${RAPPEND}" =~ "aend" ]]; then
      DATAHEX=${DATAHEX}${TFS}
      SIZE=$(( $SIZE + ${#TFS} / 2 ))
   fi
   if [[ "${RAPPEND}" =~ "lanow" ]]; then
      LANOW=1
   fi

   while [[ $i -lt $(( $SIZE * 2 - 6 )) ]]
   do
      START=$(( $i + 6 ))
      END=$START
      j=$START
      FOUND=0
      RESULT=0
      while [[ $FOUND -eq 0 ]]
      do 
         #echo "i: $i, j: $j, size: $SIZE"
         # Search
         SEARCH=${DATAHEX:$j:$PERF}
         #echo $SEARCH
         if [[ "$SEARCH" =~ "${TFS}" ]]; then
            RESULT=${DATAHEX:$START:$(( $j + $PERF ))}
            RESULT=${RESULT%%${TFS}*}
            #echo $RESULT
            FOUND=1
         fi
         j=$(( $j + ${PERF} - 4 ))
      done
      # Pick language
      LANF=$(( $INDEX % $NBLAN ))
      if [[ $LANF == ${LAN} ]] || [[ $LANOW -eq 1 ]]
      then
         NAME=`echo "${RESULT}" | xxd -r -p | sed "${SED}"`
         echo "$MM:$NAME"
         MM=$(( $MM + 1 ))
      fi
      INDEX=$(( $INDEX + 1 ))
      i=$(( $i + ${#RESULT} + 6 ))
   done
CHUNK=$(( $CHUNK - 1 ))
EXTRACT=$(( $EXTRACT + 1 ))
done
